import asyncio
import websockets
import uuid
import functools
import math
import numbers
import logging as _logging
import typing
from .package import Package

default_loop = asyncio.get_event_loop()
log = _logging.getLogger(__name__)


class NotConnectedError(Exception):
    """The :py:class:`royalnet.network.RoyalnetLink` is not connected to a :py:class:`royalnet.network.RoyalnetServer`."""


class NotIdentifiedError(Exception):
    """The :py:class:`royalnet.network.RoyalnetLink` has not identified yet to a :py:class:`royalnet.network.RoyalnetServer`."""


class ConnectionClosedError(Exception):
    """The :py:class:`royalnet.network.RoyalnetLink`'s connection was closed unexpectedly. The link can't be used anymore."""


class InvalidServerResponseError(Exception):
    """The :py:class:`royalnet.network.RoyalnetServer` sent invalid data to the :py:class:`royalnet.network.RoyalnetLink`."""


class NetworkError(Exception):
    def __init__(self, error_data: dict, *args):
        super().__init__(*args)
        self.error_data: dict = error_data


class PendingRequest:
    def __init__(self, *, loop=default_loop):
        self.event: asyncio.Event = asyncio.Event(loop=loop)
        self.data: typing.Optional[dict] = None

    def __repr__(self):
        if self.event.is_set():
            return f"<PendingRequest: {self.data.__class__.__name__}>"
        return f"<PendingRequest>"

    def set(self, data):
        self.data = data
        self.event.set()


def requires_connection(func):
    @functools.wraps(func)
    async def new_func(self, *args, **kwargs):
        await self.connect_event.wait()
        return await func(self, *args, **kwargs)
    return new_func


def requires_identification(func):
    @functools.wraps(func)
    async def new_func(self, *args, **kwargs):
        await self.identify_event.wait()
        return await func(self, *args, **kwargs)
    return new_func


class RoyalnetLink:
    def __init__(self, master_uri: str, secret: str, link_type: str, request_handler, *,
                 loop: asyncio.AbstractEventLoop = default_loop):
        assert ":" not in link_type
        self.master_uri: str = master_uri
        self.link_type: str = link_type
        self.nid: str = str(uuid.uuid4())
        self.secret: str = secret
        self.websocket: typing.Optional[websockets.WebSocketClientProtocol] = None
        self.request_handler = request_handler
        self._pending_requests: typing.Dict[str, PendingRequest] = {}
        self._loop: asyncio.AbstractEventLoop = loop
        self.error_event: asyncio.Event = asyncio.Event(loop=self._loop)
        self.connect_event: asyncio.Event = asyncio.Event(loop=self._loop)
        self.identify_event: asyncio.Event = asyncio.Event(loop=self._loop)

    async def connect(self):
        """Connect to the :py:class:`royalnet.network.RoyalnetServer` at ``self.master_uri``."""
        log.info(f"Connecting to {self.master_uri}...")
        self.websocket = await websockets.connect(self.master_uri, loop=self._loop)
        self.connect_event.set()
        log.info(f"Connected!")

    @requires_connection
    async def receive(self) -> Package:
        """Recieve a :py:class:`Package` from the :py:class:`royalnet.network.RoyalnetServer`.

        Raises:
            :py:exc:`royalnet.network.royalnetlink.ConnectionClosedError` if the connection closes."""
        try:
            jbytes: bytes = await self.websocket.recv()
            package: Package = Package.from_json_bytes(jbytes)
        except websockets.ConnectionClosed:
            self.error_event.set()
            self.connect_event.clear()
            self.identify_event.clear()
            log.info(f"Connection to {self.master_uri} was closed.")
            # What to do now? Let's just reraise.
            raise ConnectionClosedError()
        if self.identify_event.is_set() and package.destination != self.nid:
            raise InvalidServerResponseError("Package is not addressed to this RoyalnetLink.")
        log.debug(f"Received package: {package}")
        return package

    @requires_connection
    async def identify(self) -> None:
        log.info(f"Identifying to {self.master_uri}...")
        await self.websocket.send(f"Identify {self.nid}:{self.link_type}:{self.secret}")
        response: Package = await self.receive()
        if not response.source == "<server>":
            raise InvalidServerResponseError("Received a non-service package before identification.")
        if "type" not in response.data:
            raise InvalidServerResponseError("Missing 'type' in response data")
        if response.data["type"] == "error":
            raise ConnectionClosedError(f"Identification error: {response.data['type']}")
        assert response.data["type"] == "success"
        self.identify_event.set()
        log.info(f"Identified successfully!")

    @requires_identification
    async def send(self, package: Package):
        await self.websocket.send(package.to_json_bytes())
        log.debug(f"Sent package: {package}")

    @requires_identification
    async def request(self, message, destination):
        package = Package(message, source=self.nid, destination=destination)
        request = PendingRequest(loop=self._loop)
        self._pending_requests[package.source_conv_id] = request
        await self.send(package)
        log.debug(f"Sent request: {message} -> {destination}")
        await request.event.wait()
        response: dict = request.data
        log.debug(f"Received response: {request} -> {response}")
        return response

    async def run(self, loops: numbers.Real = math.inf):
        """Blockingly run the Link."""
        log.debug(f"Running main client loop for {self.nid}.")
        if self.error_event.is_set():
            raise ConnectionClosedError("RoyalnetLinks can't be rerun after an error.")
        while loops:
            loops -= 1
            if not self.connect_event.is_set():
                await self.connect()
            if not self.identify_event.is_set():
                await self.identify()
            package: Package = await self.receive()
            # Package is a response
            if package.destination_conv_id in self._pending_requests:
                request = self._pending_requests[package.destination_conv_id]
                request.set(package.data)
                continue
            # Package is a request
            assert isinstance(package, Package)
            log.debug(f"Received request {package.source_conv_id}: {package}")
            response = await self.request_handler(package.data)
            response_package: Package = package.reply(response)
            await self.send(response_package)
            log.debug(f"Replied to request {response_package.source_conv_id}: {response_package}")



from sqlalchemy import Column, \
                       Integer, \
                       Text, \
                       DateTime, \
                       ForeignKey
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declared_attr
# noinspection PyUnresolvedReferences
from .wikipages import WikiPage
# noinspection PyUnresolvedReferences
from .royals import Royal


class WikiRevision:
    """A wiki page revision.

    Warning:
        Requires PostgreSQL!"""
    __tablename__ = "wikirevisions"

    @declared_attr
    def revision_id(self):
        return Column(UUID(as_uuid=True), primary_key=True)

    @declared_attr
    def page_id(self):
        return Column(UUID(as_uuid=True), ForeignKey("wikipages.page_id"), nullable=False)

    @declared_attr
    def page(self):
        return relationship("WikiPage", foreign_keys=self.page_id, backref="revisions")

    @declared_attr
    def author_id(self):
        return Column(Integer, ForeignKey("royals.uid"), nullable=False)

    @declared_attr
    def author(self):
        return relationship("Royal", foreign_keys=self.author_id, backref="wiki_contributions")

    @declared_attr
    def timestamp(self):
        return Column(DateTime, nullable=False)

    @declared_attr
    def reason(self):
        return Column(Text)

    @declared_attr
    def diff(self):
        return Column(Text)



import typing
from ..error import UnsupportedError
if typing.TYPE_CHECKING:
    from .call import Call
    from ..utils import NetworkHandler


class Command:
    """The base class from which all commands should inherit.
    
    Attributes:
        command_name: The name of the command. To have ``/example`` on Telegram, the name should be ``example``.
        command_description: A small description of the command, to be displayed when the command is being autocompleted.
        command_syntax: The syntax of the command, to be displayed when a :py:exc:`royalnet.error.InvalidInputError` is raised, in the format ``(required_arg) [optional_arg]``.
        require_alchemy_tables: A set of :py:class:`royalnet.database` tables, that must exist for this command to work.
        network_handlers: A list of :py:class:`classyalnet.utils.NetworkHandler`s that must exist for this command to work."""

    command_name: str = NotImplemented
    command_description: str = NotImplemented
    command_syntax: str = NotImplemented

    require_alchemy_tables: typing.Set = set()

    network_handlers: typing.List[typing.Type["NetworkHandler"]] = {}

    @classmethod
    async def common(cls, call: "Call"):
        raise UnsupportedError()

    @classmethod
    def network_handler_dict(cls):
        d = {}
        for network_handler in cls.network_handlers:
            d[network_handler.message_type] = network_handler
        return d



class SafeDict(dict):
    def __missing__(self, key):
        return "{" + key + "}"


def safeformat(string: str, **words: str) -> str:
    """:py:func:`str.format` something, but ignore missing keys instead of raising an error.

    Parameters:
        string: The base string to be formatted.
        words: The words to format the string with.

    Returns:
        The formatted string."""
    return string.format_map(SafeDict(**words))



import flask as f
from werkzeug.local import LocalProxy


alchemy = f.current_app.config["ALCHEMY"]


def get_alchemy_session():
    if "alchemy_session" not in f.g:
        f.g.alchemy_session = alchemy.Session()
    return f.g.alchemy_session


@f.current_app.teardown_appcontext
def teardown_alchemy_session(*_, **__):
    _alchemy_session = f.g.pop("alchemy_session", None)
    if _alchemy_session is not None:
        _alchemy_session.close()


alchemy_session = LocalProxy(get_alchemy_session)



import typing
from sqlalchemy.inspection import inspect


def relationshiplinkchain(starting_class, ending_class) -> typing.Optional[tuple]:
    """Find the path to follow to get from the starting table to the ending table."""
    inspected = set()

    def search(_mapper, chain):
        inspected.add(_mapper)
        if _mapper.class_ == ending_class:
            return chain
        relationships = _mapper.relationships
        for _relationship in set(relationships):
            if _relationship.mapper in inspected:
                continue
            result = search(_relationship.mapper, chain + (_relationship,))
            if len(result) != 0:
                return result
        return ()

    return search(inspect(starting_class), tuple())



import random
from ..utils import Command, Call


MAD = ["MADDEN MADDEN MADDEN MADDEN",
       "EA bad, praise Geraldo!",
       "Stai sfogando la tua ira sul bot!",
       "Basta, io cambio gilda!",
       "Fondiamo la RRYG!"]


class RageCommand(Command):

    command_name = "rage"
    command_description = "Arrabbiati con qualcosa, possibilmente una software house."
    command_syntax = ""

    @classmethod
    async def common(cls, call: Call):
        await call.reply(f"😠 {random.sample(MAD, 1)[0]}")



import discord
import asyncio
import typing
import logging as _logging
from .generic import GenericBot
from ..commands import NullCommand
from ..utils import asyncify, Call, Command, discord_escape
from ..error import UnregisteredError, NoneFoundError, TooManyFoundError, InvalidConfigError, RoyalnetResponseError
from ..network import RoyalnetConfig, Request, ResponseSuccess, ResponseError
from ..database import DatabaseConfig
from ..audio import PlayMode, Playlist, RoyalPCMAudio

loop = asyncio.get_event_loop()
log = _logging.getLogger(__name__)

# TODO: Load the opus library
if not discord.opus.is_loaded():
    log.error("Opus is not loaded. Weird behaviour might emerge.")


class DiscordConfig:
    """The specific configuration to be used for :py:class:`royalnet.bots.DiscordBot`."""
    def __init__(self, token: str):
        self.token = token


class DiscordBot(GenericBot):
    """A bot that connects to `Discord <https://discordapp.com/>`_."""

    interface_name = "discord"

    def _init_voice(self):
        """Initialize the variables needed for the connection to voice chat."""
        log.debug(f"Creating music_data dict")
        self.music_data: typing.Dict[discord.Guild, PlayMode] = {}

    def _call_factory(self) -> typing.Type[Call]:
        log.debug(f"Creating DiscordCall")

        # noinspection PyMethodParameters
        class DiscordCall(Call):
            interface_name = self.interface_name
            interface_obj = self
            interface_prefix = "!"

            alchemy = self.alchemy

            async def reply(call, text: str):
                # TODO: don't escape characters inside [c][/c] blocks
                await call.channel.send(discord_escape(text))

            async def net_request(call, request: Request, destination: str) -> dict:
                if self.network is None:
                    raise InvalidConfigError("Royalnet is not enabled on this bot")
                response_dict: dict = await self.network.request(request.to_dict(), destination)
                if "type" not in response_dict:
                    raise RoyalnetResponseError("Response is missing a type")
                elif response_dict["type"] == "ResponseSuccess":
                    response: typing.Union[ResponseSuccess, ResponseError] = ResponseSuccess.from_dict(response_dict)
                elif response_dict["type"] == "ResponseError":
                    response = ResponseError.from_dict(response_dict)
                else:
                    raise RoyalnetResponseError("Response type is unknown")
                response.raise_on_error()
                return response.data

            async def get_author(call, error_if_none=False):
                message: discord.Message = call.kwargs["message"]
                user: discord.Member = message.author
                query = call.session.query(self.master_table)
                for link in self.identity_chain:
                    query = query.join(link.mapper.class_)
                query = query.filter(self.identity_column == user.id)
                result = await asyncify(query.one_or_none)
                if result is None and error_if_none:
                    raise UnregisteredError("Author is not registered")
                return result

        return DiscordCall

    def _bot_factory(self) -> typing.Type[discord.Client]:
        """Create a custom DiscordClient class inheriting from :py:class:`discord.Client`."""
        log.debug(f"Creating DiscordClient")

        # noinspection PyMethodParameters
        class DiscordClient(discord.Client):
            async def vc_connect_or_move(cli, channel: discord.VoiceChannel):
                # Connect to voice chat
                try:
                    await channel.connect()
                    log.debug(f"Connecting to Voice in {channel}")
                except discord.errors.ClientException:
                    # Move to the selected channel, instead of connecting
                    # noinspection PyUnusedLocal
                    for voice_client in cli.voice_clients:
                        voice_client: discord.VoiceClient
                        if voice_client.guild != channel.guild:
                            continue
                        await voice_client.move_to(channel)
                        log.debug(f"Moved {voice_client} to {channel}")
                # Create a music_data entry, if it doesn't exist; default is a Playlist
                if not self.music_data.get(channel.guild):
                    log.debug(f"Creating music_data for {channel.guild}")
                    self.music_data[channel.guild] = Playlist()

            @staticmethod
            async def on_message(message: discord.Message):
                text = message.content
                # Skip non-text messages
                if not text:
                    return
                # Skip bot messages
                author: typing.Union[discord.User] = message.author
                if author.bot:
                    return
                # Find and clean parameters
                command_text, *parameters = text.split(" ")
                # Don't use a case-sensitive command name
                command_name = command_text.lower()
                # Call the command
                await self.call(command_name, message.channel, parameters, message=message)

            async def on_ready(cli):
                log.debug("Connection successful, client is ready")
                await cli.change_presence(status=discord.Status.online)

            def find_guild_by_name(cli, name: str) -> discord.Guild:
                """Find the :py:class:`discord.Guild` with the specified name. Case-insensitive.

                Raises:
                     :py:exc:`NoneFoundError` if no channels are found.
                     :py:exc:`TooManyFoundError` if more than one is found."""
                all_guilds: typing.List[discord.Guild] = cli.guilds
                matching_channels: typing.List[discord.Guild] = []
                for guild in all_guilds:
                    if guild.name.lower() == name.lower():
                        matching_channels.append(guild)
                if len(matching_channels) == 0:
                    raise NoneFoundError("No channels were found")
                elif len(matching_channels) > 1:
                    raise TooManyFoundError("Too many channels were found")
                return matching_channels[0]

            def find_channel_by_name(cli,
                                     name: str,
                                     guild: typing.Optional[discord.Guild] = None) -> discord.abc.GuildChannel:
                """Find the :py:class:`TextChannel`, :py:class:`VoiceChannel` or :py:class:`CategoryChannel` with the specified name. Case-insensitive.
                Guild is optional, but the method will raise a :py:exc:`TooManyFoundError` if none is specified and there is more than one channel with the same name.
                Will also raise a :py:exc:`NoneFoundError` if no channels are found."""
                if guild is not None:
                    all_channels = guild.channels
                else:
                    all_channels: typing.List[discord.abc.GuildChannel] = cli.get_all_channels()
                matching_channels: typing.List[discord.abc.GuildChannel] = []
                for channel in all_channels:
                    if not (isinstance(channel, discord.TextChannel)
                            or isinstance(channel, discord.VoiceChannel)
                            or isinstance(channel, discord.CategoryChannel)):
                        continue
                    if channel.name.lower() == name.lower():
                        matching_channels.append(channel)
                if len(matching_channels) == 0:
                    raise NoneFoundError("No channels were found")
                elif len(matching_channels) > 1:
                    raise TooManyFoundError("Too many channels were found")
                return matching_channels[0]

            def find_voice_client_by_guild(cli, guild: discord.Guild):
                """Find the :py:class:`discord.VoiceClient` belonging to a specific :py:class:`discord.Guild`.

                Raises:
                     :py:exc:`NoneFoundError` if the :py:class:`discord.Guild` currently has no :py:class:`discord.VoiceClient`."""
                for voice_client in cli.voice_clients:
                    if voice_client.guild == guild:
                        return voice_client
                raise NoneFoundError("No voice clients found")

        return DiscordClient

    def _init_client(self):
        """Create an instance of the DiscordClient class created in :py:func:`royalnet.bots.DiscordBot._bot_factory`."""
        log.debug(f"Creating DiscordClient instance")
        self._Client = self._bot_factory()
        self.client = self._Client()

    def __init__(self, *,
                 discord_config: DiscordConfig,
                 royalnet_config: typing.Optional[RoyalnetConfig] = None,
                 database_config: typing.Optional[DatabaseConfig] = None,
                 command_prefix: str = "!",
                 commands: typing.List[typing.Type[Command]] = None,
                 missing_command: typing.Type[Command] = NullCommand,
                 error_command: typing.Type[Command] = NullCommand):
        super().__init__(royalnet_config=royalnet_config,
                         database_config=database_config,
                         command_prefix=command_prefix,
                         commands=commands,
                         missing_command=missing_command,
                         error_command=error_command)
        self._discord_config = discord_config
        self._init_client()
        self._init_voice()

    async def run(self):
        """Login to Discord, then run the bot."""
        log.debug(f"Logging in to Discord")
        await self.client.login(self._discord_config.token)
        log.debug(f"Connecting to Discord")
        await self.client.connect()
        # TODO: how to stop?

    async def add_to_music_data(self, audio_sources: typing.List[RoyalPCMAudio], guild: discord.Guild):
        """Add a file to the corresponding music_data object."""
        guild_music_data = self.music_data[guild]
        for audio_source in audio_sources:
            log.debug(f"Adding {audio_source} to music_data")
            guild_music_data.add(audio_source)
        if guild_music_data.now_playing is None:
            await self.advance_music_data(guild)

    async def advance_music_data(self, guild: discord.Guild):
        """Try to play the next song, while it exists. Otherwise, just return."""
        guild_music_data = self.music_data[guild]
        voice_client = self.client.find_voice_client_by_guild(guild)
        next_source: RoyalPCMAudio = await guild_music_data.next()
        await self.update_activity_with_source_title()
        if next_source is None:
            log.debug(f"Ending playback chain")
            return

        def advance(error=None):
            if error:
                raise Exception(f"Error while advancing music_data: {error}")
            loop.create_task(self.advance_music_data(guild))

        log.debug(f"Starting playback of {next_source}")
        voice_client.play(next_source, after=advance)

    async def update_activity_with_source_title(self):
        """Change the bot's presence (using :py:func:`discord.Client.change_presence`) to match the current listening status.

        If multiple guilds are using the bot, the bot will always have an empty presence."""
        if len(self.music_data) != 1:
            # Multiple guilds are using the bot, do not display anything
            log.debug(f"Updating current Activity: setting to None, as multiple guilds are using the bot")
            await self.client.change_presence(status=discord.Status.online)
            return
        # FIXME: PyCharm faulty inspection?
        # noinspection PyUnresolvedReferences
        play_mode: PlayMode = list(self.music_data.items())[0][1]
        now_playing = play_mode.now_playing
        if now_playing is None:
            # No songs are playing now
            log.debug(f"Updating current Activity: setting to None, as nothing is currently being played")
            await self.client.change_presence(status=discord.Status.online)
            return
        log.debug(f"Updating current Activity: listening to {now_playing.rpf.info.title}")
        await self.client.change_presence(activity=discord.Activity(name=now_playing.rpf.info.title,
                                                                    type=discord.ActivityType.listening),
                                          status=discord.Status.online)



import typing
import re


def andformat(l: typing.List[str], middle=", ", final=" and ") -> str:
    """Convert a :py:class:`list` to a :py:class:`str` by adding ``final`` between the last two elements and ``middle`` between the others.

    Parameters:
        l: the input :py:class:`list`.
        middle: the :py:class:`str` to be added between the middle elements.
        final: the :py:class:`str` to be added between the last two elements.

    Returns:
        The resulting :py:class:`str`."""
    result = ""
    for index, item in enumerate(l):
        result += item
        if index == len(l) - 2:
            result += final
        elif index != len(l) - 1:
            result += middle
    return result


def plusformat(i: int) -> str:
    """Convert an :py:class:`int` to a :py:class:`str`, prepending a ``+`` if it's greater than 0.

    Parameters:
        i: the :py:class:`int` to convert.

    Returns:
        The resulting :py:class:`str`."""
    if i >= 0:
        return f"+{i}"
    return str(i)


def fileformat(string: str) -> str:
    """Ensure a string can be used as a filename by replacing all non-word characters with underscores.
    
    Parameters:
        string: the input string.
    
    Returns:
        A valid filename string."""
    return re.sub(r"\W", "_", string)


def ytdldateformat(string: typing.Optional[str], separator: str = "-") -> str:
    """Convert the weird date string returned by ``youtube-dl`` into the ``YYYY-MM-DD`` format.
    
    Parameters:
        string: the input string, in the ``YYYYMMDD`` format.
        separator: the string to add between the years, the months and the days. Defaults to ``-``.
        
    Returns:
        The resulting string, in the format ``YYYY-MM-DD`` format."""
    if string is None:
        return ""
    return f"{string[0:4]}{separator}{string[4:6]}{separator}{string[6:8]}"


def numberemojiformat(l: typing.List[str]) -> str:
    number_emojis = ["1️⃣", "2️⃣", "3️⃣", "4️⃣", "5️⃣", "6️⃣", "7️⃣", "8️⃣", "9️⃣", "🔟"]
    extra_emoji = "*️⃣"
    result = ""
    for index, element in enumerate(l):
        try:
            result += f"{number_emojis[index]} {element}\n"
        except IndexError:
            result += f"{extra_emoji} {element}\n"
    return result



"""Various bot interfaces, and a generic class to create new ones."""

from .generic import GenericBot
from .telegram import TelegramBot, TelegramConfig
from .discord import DiscordBot, DiscordConfig

__all__ = ["TelegramBot", "TelegramConfig", "DiscordBot", "DiscordConfig", "GenericBot"]



import flask as f
import hashlib
import hmac
import datetime
from ... import Royalprint
from ....database.tables import Royal, Telegram


bp = Royalprint("tglogin", __name__, url_prefix="/login/telegram", required_tables={Royal, Telegram},
                template_folder="templates")


@bp.route("/")
def tglogin_index():
    f.session.pop("royal", None)
    return f.render_template("tglogin_index.html")


@bp.route("/done")
def tglogin_done():
    from ...alchemyhandler import alchemy, alchemy_session
    data_check_string = ""
    for field in sorted(list(f.request.args)):
        if field == "hash":
            continue
        data_check_string += f"{field}={f.request.args[field]}\n"
    data_check_string = data_check_string.rstrip("\n")
    data_check = bytes(data_check_string, encoding="ascii")
    secret_key = hashlib.sha256(bytes(f.current_app.config["TG_AK"], encoding="ascii")).digest()
    hex_data = hmac.new(key=secret_key, msg=data_check, digestmod="sha256").hexdigest()
    if hex_data != f.request.args["hash"]:
        return "Invalid authentication", 403
    tg_user = alchemy_session.query(alchemy.Telegram).filter(alchemy.Telegram.tg_id == f.request.args["id"]).one_or_none()
    if tg_user is None:
        return "No such telegram", 404
    royal_user = tg_user.royal
    f.session["royal"] = {
        "uid": royal_user.uid,
        "username": royal_user.username,
        "avatar": royal_user.avatar,
        "role": royal_user.role
    }
    f.session["login_date"] = datetime.datetime.now()
    return f.render_template("tglogin_success.html")



from sqlalchemy import Column, \
                       String, \
                       Integer, \
                       ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declared_attr
# noinspection PyUnresolvedReferences
from .royals import Royal
# noinspection PyUnresolvedReferences
from .keygroups import Keygroup


class ActiveKvGroup:
    __tablename__ = "activekvgroups"

    @declared_attr
    def royal_id(self):
        return Column(Integer, ForeignKey("royals.uid"), primary_key=True)

    @declared_attr
    def group_name(self):
        return Column(String, ForeignKey("keygroups.group_name"), nullable=False)

    @declared_attr
    def royal(self):
        return relationship("Royal", backref="active_kv_group")

    @declared_attr
    def group(self):
        return relationship("Keygroup", backref="users_with_this_active")

    def __repr__(self):
        return f"<ActiveKvGroup royal={self.royal} group={self.group_name}>"



import typing
import discord
from ..network import Request, ResponseSuccess
from ..utils import Command, Call, NetworkHandler
from ..error import TooManyFoundError, NoneFoundError
if typing.TYPE_CHECKING:
    from ..bots import DiscordBot


class PauseNH(NetworkHandler):
    message_type = "music_pause"

    # noinspection PyProtectedMember
    @classmethod
    async def discord(cls, bot: "DiscordBot", data: dict):
        # Find the matching guild
        if data["guild_name"]:
            guild = bot.client.find_guild_by_name(data["guild_name"])
        else:
            if len(bot.music_data) == 0:
                raise NoneFoundError("No voice clients active")
            if len(bot.music_data) > 1:
                raise TooManyFoundError("Multiple guilds found")
            guild = list(bot.music_data)[0]
        # Set the currently playing source as ended
        voice_client: discord.VoiceClient = bot.client.find_voice_client_by_guild(guild)
        if not (voice_client.is_playing() or voice_client.is_paused()):
            raise NoneFoundError("Nothing to pause")
        # Toggle pause
        resume = voice_client._player.is_paused()
        if resume:
            voice_client._player.resume()
        else:
            voice_client._player.pause()
        return ResponseSuccess({"resume": resume})


class PauseCommand(Command):

    command_name = "pause"
    command_description = "Mette in pausa o riprende la riproduzione della canzone attuale."
    command_syntax = "[ [guild] ]"

    network_handlers = [PauseNH]

    @classmethod
    async def common(cls, call: Call):
        guild, = call.args.match(r"(?:\[(.+)])?")
        response = await call.net_request(Request("music_pause", {"guild_name": guild}), "discord")
        if response["resume"]:
            await call.reply(f"▶️ Riproduzione ripresa.")
        else:
            await call.reply(f"⏸ Riproduzione messa in pausa.")



import typing

def cdj(class_: typing.Any) -> dict:
    """Return a dict of the class attributes without the ``__module__``, ``__dict__``, ``__weakref__`` and ``__doc__`` keys, to be used while generating dynamically SQLAlchemy declarative table classes.
    
    Parameters:
        class_: The object that you want to dict-ify. 

    Returns:
        The class dict.
        
    Warning:
        You can't dict-ify classes with ``__slots__``!"""
    d = dict(class_.__dict__)
    del d["__module__"]
    del d["__dict__"]
    del d["__weakref__"]
    del d["__doc__"]
    return d



import datetime
import dateparser
import typing
from ..utils import Command, Call, sleep_until


class ReminderCommand(Command):

    command_name = "reminder"
    command_description = "Ripete quello che gli avevi chiesto dopo un po' di tempo."
    command_syntax = "[ (data) ] (testo)"

    @classmethod
    async def common(cls, call: Call):
        date_str, reminder_text = call.args.match(r"\[ *(.+?) *] *(.+?) *$")
        try:
            date: typing.Optional[datetime.datetime] = dateparser.parse(date_str)
        except OverflowError:
            date = None
        if date is None:
            await call.reply("⚠️ La data che hai inserito non è valida.")
            return
        await call.reply(f"✅ Promemoria impostato per [b]{date.strftime('%Y-%m-%d %H:%M:%S')}[/b]")
        await sleep_until(date)
        await call.reply(f"❗️ Promemoria: [b]{reminder_text}[/b]")



import flask as f
from ... import Royalprint
from ....database.tables import Royal


bp = Royalprint("testing", __name__, url_prefix="/testing", required_tables={Royal})


@bp.route("/listroyals")
def testing_listroyals():
    from ...alchemyhandler import alchemy, alchemy_session
    royals = alchemy_session.query(alchemy.Royal).all()
    return f'<body><script type="text/plain" style="display: block;">{repr(royals)}</script></body>'



import typing
import flask as f
import os
from ..database import Alchemy
from .royalprint import Royalprint


def create_app(config_obj: typing.Type, blueprints: typing.List[Royalprint]):
    app = f.Flask(__name__)
    app.config.from_object(config_obj)
    app.secret_key = os.environ["SECRET_KEY"]
    required_tables = set()
    for blueprint in blueprints:
        required_tables = required_tables.union(blueprint.required_tables)
        app.register_blueprint(blueprint)
    app.config["ALCHEMY"] = Alchemy(app.config["DB_PATH"], required_tables)
    return app



from sqlalchemy import Column, \
                       String, \
                       ForeignKey
from sqlalchemy.ext.declarative import declared_attr


class Keygroup:
    __tablename__ = "keygroups"

    @declared_attr
    def group_name(self):
        return Column(String, ForeignKey("keygroups.group_name"), primary_key=True)

    def __repr__(self):
        return f"<Keygroup {self.group_name}>"



import asyncio
from ..utils import Command, Call, asyncify
from ..audio import YtdlInfo


class VideoinfoCommand(Command):

    command_name = "videoinfo"
    command_description = "Visualizza le informazioni di un video."
    command_syntax = "(url)"

    @classmethod
    async def common(cls, call: Call):
        url = call.args[0]
        info_list = await asyncify(YtdlInfo.create_from_url, url)
        for info in info_list:
            info_dict = info.__dict__
            message = f"🔍 Dati di [b]{info}[/b]:\n"
            for key in info_dict:
                # Skip description
                if key == "description":
                    continue
                # Skip formats
                if key == "formats":
                    continue
                if key == "requested_formats":
                    continue
                # Skip subtitles
                if key == "subtitles":
                    continue
                # Skip empty keys
                if info_dict[key] is None:
                    continue
                message += f"[c]{key}[/c]: {info_dict[key]}\n"
            await call.reply(message)
            await asyncio.sleep(0.2)



from sqlalchemy import Column, \
                       Text, \
                       String
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.ext.declarative import declared_attr


class WikiPage:
    """Wiki page properties.

    Warning:
        Requires PostgreSQL!"""
    __tablename__ = "wikipages"

    @declared_attr
    def page_id(self):
        return Column(UUID(as_uuid=True), primary_key=True)

    @declared_attr
    def title(self):
        return Column(String, nullable=False)

    @declared_attr
    def content(self):
        return Column(Text)



import typing
import discord
import asyncio
from ..utils import Command, Call, NetworkHandler, andformat
from ..network import Request, ResponseSuccess
from ..error import NoneFoundError, TooManyFoundError
if typing.TYPE_CHECKING:
    from ..bots import DiscordBot


loop = asyncio.get_event_loop()


class CvNH(NetworkHandler):
    message_type = "discord_cv"

    @classmethod
    async def discord(cls, bot: "DiscordBot", data: dict):
        # Find the matching guild
        if data["guild_name"]:
            guild: discord.Guild = bot.client.find_guild_by_name(data["guild_name"])
        else:
            if len(bot.client.guilds) == 0:
                raise NoneFoundError("No guilds found")
            if len(bot.client.guilds) > 1:
                raise TooManyFoundError("Multiple guilds found")
            guild = list(bot.client.guilds)[0]
        # Edit the message, sorted by channel
        discord_members = list(guild.members)
        channels = {0: None}
        members_in_channels = {0: []}
        message = ""
        # Find all the channels
        for member in discord_members:
            if member.voice is not None:
                channel = members_in_channels.get(member.voice.channel.id)
                if channel is None:
                    members_in_channels[member.voice.channel.id] = list()
                    channel = members_in_channels[member.voice.channel.id]
                    channels[member.voice.channel.id] = member.voice.channel
                channel.append(member)
            else:
                members_in_channels[0].append(member)
        # Edit the message, sorted by channel
        for channel in sorted(channels, key=lambda c: -c):
            members_in_channels[channel].sort(key=lambda x: x.nick if x.nick is not None else x.name)
            if channel == 0:
                message += "[b]Non in chat vocale:[/b]\n"
            else:
                message += f"[b]In #{channels[channel].name}:[/b]\n"
            for member in members_in_channels[channel]:
                member: typing.Union[discord.User, discord.Member]
                # Ignore not-connected non-notable members
                if not data["full"] and channel == 0 and len(member.roles) < 2:
                    continue
                # Ignore offline members
                if member.status == discord.Status.offline and member.voice is None:
                    continue
                # Online status emoji
                if member.bot:
                    message += "🤖 "
                elif member.status == discord.Status.online:
                    message += "🔵 "
                elif member.status == discord.Status.idle:
                    message += "⚫️ "
                elif member.status == discord.Status.dnd:
                    message += "🔴 "
                elif member.status == discord.Status.offline:
                    message += "⚪️ "
                # Voice
                if channel != 0:
                    # Voice status
                    if member.voice.self_mute:
                        message += f"🔇 "
                    else:
                        message += f"🔊 "
                # Nickname
                if member.nick is not None:
                    message += f"[i]{member.nick}[/i]"
                else:
                    message += member.name
                # Game or stream
                if member.activity is not None:
                    if member.activity.type == discord.ActivityType.playing:
                        message += f" | 🎮 {member.activity.name}"
                        # Rich presence
                        try:
                            if member.activity.state is not None:
                                message += f" ({member.activity.state}" \
                                           f" | {member.activity.details})"
                        except AttributeError:
                            pass
                    elif member.activity.type == discord.ActivityType.streaming:
                        message += f" | 📡 {member.activity.url})"
                    elif member.activity.type == discord.ActivityType.listening:
                        if isinstance(member.activity, discord.Spotify):
                            if member.activity.title == member.activity.album:
                                message += f" | 🎧 {member.activity.title} ({andformat(member.activity.artists, final=' e ')})"
                            else:
                                message += f" | 🎧 {member.activity.title} ({member.activity.album} | {andformat(member.activity.artists, final=' e ')})"
                        else:
                            message += f" | 🎧 {member.activity.name}"
                    elif member.activity.type == discord.ActivityType.watching:
                        message += f" | 📺 {member.activity.name}"
                message += "\n"
            message += "\n"
        return ResponseSuccess({"response": message})


class CvCommand(Command):

    command_name = "cv"
    command_description = "Elenca le persone attualmente connesse alla chat vocale."
    command_syntax = "[guildname]"

    network_handlers = [CvNH]

    @classmethod
    async def common(cls, call: Call):
        guild_name = call.args.optional(0)
        response = await call.net_request(Request("discord_cv", {"guild_name": guild_name, "full": False}), "discord")
        await call.reply(response["response"])



import math
import random
import typing
from .royalpcmaudio import RoyalPCMAudio


class PlayMode:
    """The base class for a PlayMode, such as :py:class:`royalnet.audio.Playlist`. Inherit from this class if you want to create a custom PlayMode."""

    def __init__(self):
        """Create a new PlayMode and initialize the generator inside."""
        self.now_playing: typing.Optional[RoyalPCMAudio] = None
        self.generator: typing.AsyncGenerator = self._generate_generator()

    async def next(self) -> typing.Optional[RoyalPCMAudio]:
        """Get the next :py:class:`royalnet.audio.RoyalPCMAudio` from the list and advance it.

        Returns:
            The next :py:class:`royalnet.audio.RoyalPCMAudio`."""
        return await self.generator.__anext__()

    def videos_left(self) -> typing.Union[int, float]:
        """Return the number of videos left in the PlayMode.

        Returns:
            Usually a :py:class:`int`, but may return also :py:obj:`math.inf` if the PlayMode is infinite."""
        raise NotImplementedError()

    async def _generate_generator(self):
        """Factory function for an async generator that changes the ``now_playing`` property either to a :py:class:`discord.audio.RoyalPCMAudio` or to ``None``, then yields the value it changed it to.

        Yields:
            The :py:class:`royalnet.audio.RoyalPCMAudio` to be played next."""
        raise NotImplementedError()
        # This is needed to make the coroutine an async generator
        # noinspection PyUnreachableCode
        yield NotImplemented

    def add(self, item: RoyalPCMAudio) -> None:
        """Add a new :py:class:`royalnet.audio.RoyalPCMAudio` to the PlayMode.

        Args:
            item: The item to add to the PlayMode."""
        raise NotImplementedError()

    def delete(self) -> None:
        """Delete all :py:class:`royalnet.audio.RoyalPCMAudio` contained inside this PlayMode."""
        raise NotImplementedError()

    def queue_preview(self) -> typing.List[RoyalPCMAudio]:
        """Display all the videos in the PlayMode as a list, if possible.

        To be used with `queue` commands, for example.

        Raises:
            NotImplementedError: If a preview can't be generated.

        Returns:
            A list of videos contained in the queue."""
        raise NotImplementedError()


class Playlist(PlayMode):
    """A video list. :py:class:`royalnet.audio.RoyalPCMAudio` played are removed from the list."""

    def __init__(self, starting_list: typing.List[RoyalPCMAudio] = None):
        """Create a new Playlist.

        Args:
            starting_list: A list of items with which the Playlist will be created."""
        super().__init__()
        if starting_list is None:
            starting_list = []
        self.list: typing.List[RoyalPCMAudio] = starting_list

    def videos_left(self) -> typing.Union[int, float]:
        return len(self.list)

    async def _generate_generator(self):
        while True:
            try:
                next_video = self.list.pop(0)
            except IndexError:
                self.now_playing = None
            else:
                self.now_playing = next_video
            yield self.now_playing
            if self.now_playing is not None:
                self.now_playing.delete()

    def add(self, item) -> None:
        self.list.append(item)

    def delete(self) -> None:
        if self.now_playing is not None:
            self.now_playing.delete()
        while self.list:
            self.list.pop(0).delete()

    def queue_preview(self) -> typing.List[RoyalPCMAudio]:
        return self.list


class Pool(PlayMode):
    """A :py:class:`royalnet.audio.RoyalPCMAudio` pool. :py:class:`royalnet.audio.RoyalPCMAudio` are selected in random order and are not repeated until every song has been played at least once."""

    def __init__(self, starting_pool: typing.List[RoyalPCMAudio] = None):
        """Create a new Pool.

        Args:
            starting_pool: A list of items the Pool will be created from."""
        super().__init__()
        if starting_pool is None:
            starting_pool = []
        self.pool: typing.List[RoyalPCMAudio] = starting_pool
        self._pool_copy: typing.List[RoyalPCMAudio] = []

    def videos_left(self) -> typing.Union[int, float]:
        return math.inf

    async def _generate_generator(self):
        while True:
            if not self.pool:
                self.now_playing = None
                yield None
                continue
            self._pool_copy = self.pool.copy()
            random.shuffle(self._pool_copy)
            while self._pool_copy:
                next_video = self._pool_copy.pop(0)
                self.now_playing = next_video
                yield next_video

    def add(self, item) -> None:
        self.pool.append(item)
        self._pool_copy.append(item)
        random.shuffle(self._pool_copy)

    def delete(self) -> None:
        for item in self.pool:
            item.delete()
        self.pool = None
        self._pool_copy = None

    def queue_preview(self) -> typing.List[RoyalPCMAudio]:
        preview_pool = self.pool.copy()
        random.shuffle(preview_pool)
        return preview_pool


import logging as _logging
from ..utils import Command, Call
from ..error import *

log = _logging.getLogger(__name__)


class ErrorHandlerCommand(Command):

    command_name = "error_handler"
    command_description = "Gestisce gli errori causati dagli altri comandi."
    command_syntax = ""

    @classmethod
    async def common(cls, call: Call):
        exception: Exception = call.kwargs["exception"]
        if isinstance(exception, NoneFoundError):
            await call.reply(f"⚠️ L'elemento richiesto non è stato trovato.\n[p]{exception}[/p]")
            return
        if isinstance(exception, TooManyFoundError):
            await call.reply(f"⚠️ La richiesta effettuata è ambigua, pertanto è stata annullata.\n[p]{exception}[/p]")
            return
        if isinstance(exception, UnregisteredError):
            await call.reply("⚠️ Devi essere registrato a Royalnet per usare questo comando.\nUsa il comando [c]sync[/c] per registrarti!")
            return
        if isinstance(exception, UnsupportedError):
            await call.reply(f"⚠️ Il comando richiesto non è disponibile tramite l'interfaccia [c]{call.interface_name}[/c].")
            return
        if isinstance(exception, InvalidInputError):
            command = call.kwargs["previous_command"]
            await call.reply(f"⚠️ Sintassi non valida.\nSintassi corretta: [c]{call.interface_prefix}{command.command_name} {command.command_syntax}[/c]")
            return
        if isinstance(exception, InvalidConfigError):
            await call.reply(f"⚠️ Il bot non è stato configurato correttamente, quindi questo comando non può essere eseguito.\n[p]{exception}[/p]")
            return
        if isinstance(exception, RoyalnetRequestError):
            await call.reply(f"⚠️ La richiesta a Royalnet ha restituito un errore:\n"
                             f"[p]{exception.error.extra_info['type']}\n"
                             f"{exception.error.extra_info['str']}[/p]")
            return
        if isinstance(exception, ExternalError):
            await call.reply(f"⚠️ Una risorsa esterna necessaria per l'esecuzione del comando non ha funzionato correttamente, quindi il comando è stato annullato.\n[p]{exception}[/p]")
            return
        if isinstance(exception, RoyalnetResponseError):
            log.warning(f"Invalid response from Royalnet - {exception.__class__.__name__}: {exception}")
            await call.reply(f"❌ La risposta ricevuta da Royalnet non è valida: [p]{exception}[/p]")
            return
        log.error(f"Unhandled exception - {exception.__class__.__name__}: {exception}")
        await call.reply(f"❌ Eccezione non gestita durante l'esecuzione del comando:\n[b]{exception.__class__.__name__}[/b]\n[p]{exception}[/p]")



import typing
import pickle
from ..network import Request, ResponseSuccess
from ..utils import Command, Call, NetworkHandler, numberemojiformat
from ..error import TooManyFoundError, NoneFoundError
if typing.TYPE_CHECKING:
    from ..bots import DiscordBot


class QueueNH(NetworkHandler):
    message_type = "music_queue"

    @classmethod
    async def discord(cls, bot: "DiscordBot", data: dict):
        # Find the matching guild
        if data["guild_name"]:
            guild = bot.client.find_guild_by_name(data["guild_name"])
        else:
            if len(bot.music_data) == 0:
                raise NoneFoundError("No voice clients active")
            if len(bot.music_data) > 1:
                raise TooManyFoundError("Multiple guilds found")
            guild = list(bot.music_data)[0]
        # Check if the guild has a PlayMode
        playmode = bot.music_data.get(guild)
        if not playmode:
            return ResponseSuccess({
                "type": None
            })
        try:
            queue = playmode.queue_preview()
        except NotImplementedError:
            return ResponseSuccess({
                "type": playmode.__class__.__name__
            })
        return ResponseSuccess({
            "type": playmode.__class__.__name__,
            "queue":
            {
                "strings": [str(element.rpf.info) for element in queue],
                "pickled_embeds": str(pickle.dumps([element.rpf.info.to_discord_embed() for element in queue]))
            }
        })


class QueueCommand(Command):

    command_name = "queue"
    command_description = "Visualizza un'anteprima della coda di riproduzione attuale."
    command_syntax = "[ [guild] ]"

    network_handlers = [QueueNH]

    @classmethod
    async def common(cls, call: Call):
        guild, = call.args.match(r"(?:\[(.+)])?")
        data = await call.net_request(Request("music_queue", {"guild_name": guild}), "discord")
        if data["type"] is None:
            await call.reply("ℹ️ Non c'è nessuna coda di riproduzione attiva al momento.")
            return
        elif "queue" not in data:
            await call.reply(f"ℹ️ La coda di riproduzione attuale ([c]{data['type']}[/c]) non permette l'anteprima.")
            return
        if data["type"] == "Playlist":
            if len(data["queue"]["strings"]) == 0:
                message = f"ℹ️ Questa [c]Playlist[/c] è vuota."
            else:
                message = f"ℹ️ Questa [c]Playlist[/c] contiene {len(data['queue']['strings'])} elementi, e i prossimi saranno:\n"
        elif data["type"] == "Pool":
            if len(data["queue"]["strings"]) == 0:
                message = f"ℹ️ Questo [c]Pool[/c] è vuoto."
            else:
                message = f"ℹ️ Questo [c]Pool[/c] contiene {len(data['queue']['strings'])} elementi, tra cui:\n"
        else:
            if len(data["queue"]["strings"]) == 0:
                message = f"ℹ️ Il PlayMode attuale, [c]{data['type']}[/c], è vuoto.\n"
            else:
                message = f"ℹ️ Il PlayMode attuale, [c]{data['type']}[/c], contiene {len(data['queue']['strings'])} elementi:\n"
        if call.interface_name == "discord":
            await call.reply(message)
            for embed in pickle.loads(eval(data["queue"]["pickled_embeds"]))[:5]:
                await call.channel.send(embed=embed)
        else:
            message += numberemojiformat(data["queue"]["strings"][:10])
            await call.reply(message)



import flask as f
import markdown2
import re
import uuid
import os
from ... import Royalprint
from ....database.tables import Royal, WikiPage, WikiRevision


tmpl_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'templates')
bp = Royalprint("wikiview", __name__, url_prefix="/wikiview", template_folder=tmpl_dir,
                required_tables={Royal, WikiPage, WikiRevision})


def prepare_page(page):
    converted_md = markdown2.markdown(page.content.replace("<", "&lt;"),
                                      extras=["spoiler", "tables", "smarty-pants", "fenced-code-blocks"])
    converted_md = re.sub(r"{https?://(?:www\.)?(?:youtube\.com/watch\?.*?&?v=|youtu.be/)([0-9A-Za-z-]+).*?}",
                          r'<div class="youtube-embed">'
                          r'   <iframe src="https://www.youtube-nocookie.com/embed/\1?rel=0&amp;showinfo=0"'
                          r'           frameborder="0"'
                          r'           allow="autoplay; encrypted-media"'
                          r'           allowfullscreen'
                          r'           width="640px"'
                          r'           height="320px">'
                          r'   </iframe>'
                          r'</div>', converted_md)
    converted_md = re.sub(r"{https?://clyp.it/([a-z0-9]+)}",
                          r'<div class="clyp-embed">'
                          r'    <iframe width="100%" height="160" src="https://clyp.it/\1/widget" frameborder="0">'
                          r'    </iframe>'
                          r'</div>', converted_md)
    return converted_md


@bp.route("/")
def wikiview_index():
    from ...alchemyhandler import alchemy, alchemy_session
    pages = alchemy_session.query(alchemy.WikiPage).all()
    return f.render_template("wikiview_index.html", pages=pages)


@bp.route("/id/<page_id>")
def wikiview_by_id(page_id: str):
    from ...alchemyhandler import alchemy, alchemy_session
    page_uuid = uuid.UUID(page_id)
    page = alchemy_session.query(alchemy.WikiPage).filter(alchemy.WikiPage.page_id == page_uuid).one_or_none()
    if page is None:
        return "No such page", 404
    parsed_content = prepare_page(page)
    return f.render_template("wikiview_page.html", page=page, parsed_content=f.Markup(parsed_content))


@bp.route("/title/<title>")
def wikiview_by_title(title: str):
    from ...alchemyhandler import alchemy, alchemy_session
    page = alchemy_session.query(alchemy.WikiPage).filter(alchemy.WikiPage.title == title).one_or_none()
    if page is None:
        return "No such page", 404
    parsed_content = prepare_page(page)
    return f.render_template("wikiview_page.html", page=page, parsed_content=f.Markup(parsed_content))



import os
from .web import create_app
from .web.blueprints import home, wikiview, tglogin


class TestConfig:
    DB_PATH = os.environ["DB_PATH"]
    TG_AK = os.environ["TG_AK"]


app = create_app(TestConfig, [home, wikiview, tglogin])


if __name__ == "__main__":
    app.run()



from .royals import Royal
from .telegram import Telegram
from .diario import Diario
from .aliases import Alias
from .activekvgroups import ActiveKvGroup
from .keyvalues import Keyvalue
from .keygroups import Keygroup
from .discord import Discord
from .wikipages import WikiPage
from .wikirevisions import WikiRevision

__all__ = ["Royal", "Telegram", "Diario", "Alias", "ActiveKvGroup", "Keyvalue", "Keygroup", "Discord", "WikiPage", "WikiRevision"]



from ..utils import Command, Call


class ColorCommand(Command):

    command_name = "color"
    command_description = "Invia un colore in chat...?"
    command_syntax = ""

    @classmethod
    async def common(cls, call: Call):
        await call.reply("""
        [i]I am sorry, unknown error occured during working with your request, Admin were notified[/i]
        """)



from .helloworld import bp as helloworld
from .testing import bp as testing
from .home import bp as home
from .wikiview import bp as wikiview
from .tglogin import bp as tglogin

__all__ = ["helloworld", "testing", "home", "wikiview", "tglogin"]



import sys
import typing
import asyncio
import logging
from ..utils import Command, NetworkHandler, Call
from ..commands import NullCommand
from ..network import RoyalnetLink, Request, Response, ResponseError, RoyalnetConfig
from ..database import Alchemy, DatabaseConfig, relationshiplinkchain


loop = asyncio.get_event_loop()
log = logging.getLogger(__name__)


class GenericBot:
    """A generic bot class, to be used as base for the other more specific classes, such as :ref:`royalnet.bots.TelegramBot` and :ref:`royalnet.bots.DiscordBot`."""
    interface_name = NotImplemented

    def _init_commands(self,
                       command_prefix: str,
                       commands: typing.List[typing.Type[Command]],
                       missing_command: typing.Type[Command],
                       error_command: typing.Type[Command]) -> None:
        """Generate the ``commands`` dictionary required to handle incoming messages, and the ``network_handlers`` dictionary required to handle incoming requests."""
        log.debug(f"Now generating commands")
        self.commands: typing.Dict[str, typing.Type[Command]] = {}
        self.network_handlers: typing.Dict[str, typing.Type[NetworkHandler]] = {}
        for command in commands:
            lower_command_name = command.command_name.lower()
            self.commands[f"{command_prefix}{lower_command_name}"] = command
            self.network_handlers = {**self.network_handlers, **command.network_handler_dict()}
        self.missing_command: typing.Type[Command] = missing_command
        self.error_command: typing.Type[Command] = error_command
        log.debug(f"Successfully generated commands")

    def _call_factory(self) -> typing.Type[Call]:
        """Create the TelegramCall class, representing a command call. It should inherit from :py:class:`royalnet.utils.Call`.

        Returns:
            The created TelegramCall class."""
        raise NotImplementedError()

    def _init_royalnet(self, royalnet_config: RoyalnetConfig):
        """Create a :py:class:`royalnet.network.RoyalnetLink`, and run it as a :py:class:`asyncio.Task`."""
        self.network: RoyalnetLink = RoyalnetLink(royalnet_config.master_uri, royalnet_config.master_secret, self.interface_name,
                                                  self._network_handler)
        log.debug(f"Running RoyalnetLink {self.network}")
        loop.create_task(self.network.run())

    async def _network_handler(self, request_dict: dict) -> dict:
        """Handle a single :py:class:`dict` received from the :py:class:`royalnet.network.RoyalnetLink`.

        Returns:
            Another :py:class:`dict`, formatted as a :py:class:`royalnet.network.Response`."""
        # Convert the dict to a Request
        try:
            request: Request = Request.from_dict(request_dict)
        except TypeError:
            log.warning(f"Invalid request received: {request_dict}")
            return ResponseError("invalid_request",
                                 f"The Request that you sent was invalid. Check extra_info to see what you sent.",
                                 extra_info={"you_sent": request_dict}).to_dict()
        log.debug(f"Received {request} from the RoyalnetLink")
        try:
            network_handler = self.network_handlers[request.handler]
        except KeyError:
            log.warning(f"Missing network_handler for {request.handler}")
            return ResponseError("no_handler", f"This Link is missing a network handler for {request.handler}.").to_dict()
        try:
            log.debug(f"Using {network_handler} as handler for {request.handler}")
            response: Response = await getattr(network_handler, self.interface_name)(self, request.data)
            return response.to_dict()
        except Exception:
            _, exc, _ = sys.exc_info()
            log.debug(f"Exception {exc} in {network_handler}")
            return ResponseError("exception_in_handler",
                                 f"An exception was raised in {network_handler} for {request.handler}. Check extra_info for details.",
                                 extra_info={
                                     "type": exc.__class__.__name__,
                                     "str": str(exc)
                                 }).to_dict()

    def _init_database(self, commands: typing.List[typing.Type[Command]], database_config: DatabaseConfig):
        """Create an :py:class:`royalnet.database.Alchemy` with the tables required by the commands. Then, find the chain that links the ``master_table`` to the ``identity_table``."""
        log.debug(f"Initializing database")
        required_tables = set()
        for command in commands:
            required_tables = required_tables.union(command.require_alchemy_tables)
        log.debug(f"Found {len(required_tables)} required tables")
        self.alchemy = Alchemy(database_config.database_uri, required_tables)
        self.master_table = self.alchemy.__getattribute__(database_config.master_table.__name__)
        self.identity_table = self.alchemy.__getattribute__(database_config.identity_table.__name__)
        self.identity_column = self.identity_table.__getattribute__(self.identity_table,
                                                                    database_config.identity_column_name)
        self.identity_chain = relationshiplinkchain(self.master_table, self.identity_table)
        log.debug(f"Identity chain is {self.identity_chain}")

    def __init__(self, *,
                 royalnet_config: typing.Optional[RoyalnetConfig] = None,
                 database_config: typing.Optional[DatabaseConfig] = None,
                 command_prefix: str,
                 commands: typing.List[typing.Type[Command]] = None,
                 missing_command: typing.Type[Command] = NullCommand,
                 error_command: typing.Type[Command] = NullCommand):
        if database_config is None:
            self.alchemy = None
            self.master_table = None
            self.identity_table = None
            self.identity_column = None
        else:
            self._init_database(commands=commands, database_config=database_config)
        if commands is None:
            commands = []
        self._init_commands(command_prefix, commands, missing_command=missing_command, error_command=error_command)
        self._Call = self._call_factory()
        if royalnet_config is None:
            self.network = None
        else:
            self._init_royalnet(royalnet_config=royalnet_config)

    async def call(self, command_name: str, channel, parameters: typing.List[str] = None, **kwargs):
        """Call the command with the specified name.

        If it doesn't exist, call ``self.missing_command``.

        If an exception is raised during the execution of the command, call ``self.error_command``."""
        log.debug(f"Trying to call {command_name}")
        if parameters is None:
            parameters = []
        try:
            command: typing.Type[Command] = self.commands[command_name]
        except KeyError:
            log.debug(f"Calling missing_command because {command_name} does not exist")
            command = self.missing_command
        try:
            await self._Call(channel, command, parameters, **kwargs).run()
        except Exception as exc:
            log.debug(f"Calling error_command because of an error in {command_name}")
            await self._Call(channel, self.error_command,
                             exception=exc,
                             previous_command=command, **kwargs).run()

    async def run(self):
        """A blocking coroutine that should make the bot start listening to commands and requests."""
        raise NotImplementedError()



import json
import uuid
import typing


class Package:
    """A Royalnet package, the data type with which a :py:class:`royalnet.network.RoyalnetLink` communicates with a :py:class:`royalnet.network.RoyalnetServer` or another link.
    Contains info about the source and the destination."""

    def __init__(self,
                 data: dict,
                 *,
                 source: str,
                 destination: str,
                 source_conv_id: typing.Optional[str] = None,
                 destination_conv_id: typing.Optional[str] = None):
        """Create a Package.

        Parameters:
            data: The data that should be sent. Usually a :py:class:`royalnet.network.Message`.
            source: The ``nid`` of the node that created this Package.
            destination: The ``link_type`` of the destination node, or alternatively, the ``nid`` of the node. Can also be the ``NULL`` value to send the message to nobody.
            source_conv_id: The conversation id of the node that created this package. Akin to the sequence number on IP packets.
            destination_conv_id: The conversation id of the node that this Package is a reply to."""
        # TODO: something is not right in these type hints. Check them.
        self.data: dict = data
        self.source: str = source
        self.source_conv_id: str = source_conv_id or str(uuid.uuid4())
        self.destination: str = destination
        self.destination_conv_id: typing.Optional[str] = destination_conv_id

    def __repr__(self):
        return f"<Package {self.source} ({self.source_conv_id}) to {self.destination} ({self.destination_conv_id}>"

    def __eq__(self, other):
        if isinstance(other, Package):
            return (self.data == other.data) and \
                   (self.source == other.source) and \
                   (self.destination == other.destination) and \
                   (self.source_conv_id == other.source_conv_id) and \
                   (self.destination_conv_id == other.destination_conv_id)
        return False

    def reply(self, data) -> "Package":
        """Reply to this Package with another Package.

        Parameters:
            data: The data that should be sent. Usually a :py:class:`royalnet.network.Message`.

        Returns:
            The reply Package."""
        return Package(data,
                       source=self.destination,
                       destination=self.source,
                       source_conv_id=self.destination_conv_id or str(uuid.uuid4()),
                       destination_conv_id=self.source_conv_id)

    @staticmethod
    def from_dict(d) -> "Package":
        """Create a Package from a dictionary."""
        if "source" not in d:
            raise ValueError("Missing source field")
        if "nid" not in d["source"]:
            raise ValueError("Missing source.nid field")
        if "conv_id" not in d["source"]:
            raise ValueError("Missing source.conv_id field")
        if "destination" not in d:
            raise ValueError("Missing destination field")
        if "nid" not in d["destination"]:
            raise ValueError("Missing destination.nid field")
        if "conv_id" not in d["destination"]:
            raise ValueError("Missing destination.conv_id field")
        if "data" not in d:
            raise ValueError("Missing data field")
        return Package(d["data"],
                       source=d["source"]["nid"],
                       destination=d["destination"]["nid"],
                       source_conv_id=d["source"]["conv_id"],
                       destination_conv_id=d["destination"]["conv_id"])

    def to_dict(self) -> dict:
        """Convert the Package into a dictionary."""
        return {
            "source": {
                "nid": self.source,
                "conv_id": self.source_conv_id
            },
            "destination": {
                "nid": self.destination,
                "conv_id": self.destination_conv_id
            },
            "data": self.data
        }

    @staticmethod
    def from_json_string(string: str) -> "Package":
        """Create a Package from a JSON string."""
        return Package.from_dict(json.loads(string))

    def to_json_string(self) -> str:
        """Convert the Package into a JSON string."""
        return json.dumps(self.to_dict())

    @staticmethod
    def from_json_bytes(b: bytes) -> "Package":
        """Create a Package from UTF8-encoded JSON bytes."""
        return Package.from_json_string(str(b, encoding="utf8"))

    def to_json_bytes(self) -> bytes:
        """Convert the Package into UTF8-encoded JSON bytes."""
        return bytes(self.to_json_string(), encoding="utf8")



from discord import AudioSource
from discord.opus import Encoder as OpusEncoder
import typing
from .royalpcmfile import RoyalPCMFile


class RoyalPCMAudio(AudioSource):
    """A :py:class:`discord.AudioSource` that keeps data in a file instead of in memory."""

    def __init__(self, rpf: "RoyalPCMFile"):
        """Create a :py:class:`discord.audio.RoyalPCMAudio` from a :py:class:`royalnet.audio.RoyalPCMFile`.

        Warning:
            Not recommended, use :py:func:`royalnet.audio.RoyalPCMAudio.create_from_url` or :py:func:`royalnet.audio.RoyalPCMAudio.create_from_ytsearch` instead."""
        self.rpf: "RoyalPCMFile" = rpf
        self._file = open(self.rpf.audio_filename, "rb")

    @staticmethod
    def create_from_url(url: str) -> typing.List["RoyalPCMAudio"]:
        """Download a file with youtube_dl and create a list of RoyalPCMAudios.

        Parameters:
            url: The url of the file to download.

        Returns:
            A :py:class:`list` of RoyalPCMAudios, each corresponding to a downloaded video."""
        rpf_list = RoyalPCMFile.create_from_url(url)
        return [RoyalPCMAudio(rpf) for rpf in rpf_list]

    @staticmethod
    def create_from_ytsearch(search: str, amount: int = 1) -> typing.List["RoyalPCMAudio"]:
        """Search a string on YouTube and download the first ``amount`` number of videos, then download those with youtube_dl and create a list of RoyalPCMAudios.

        Parameters:
            search: The string to search on YouTube.
            amount: The number of videos to download.

        Returns:
            A :py:class:`list` of RoyalPCMAudios, each corresponding to a downloaded video."""
        rpf_list = RoyalPCMFile.create_from_ytsearch(search, amount)
        return [RoyalPCMAudio(rpf) for rpf in rpf_list]

    def is_opus(self):
        """This audio file isn't Opus-encoded, but PCM-encoded.
        
        Returns:
            ``False``."""
        return False

    def read(self):
        """Reads 20ms worth of audio.
        
        If the audio is complete, then returning an empty :py:class:`bytes`-like object to signal this is the way to do so."""
        data: bytes = self._file.read(OpusEncoder.FRAME_SIZE)
        # If the file was externally closed, it means it was deleted
        if self._file.closed:
            return b""
        if len(data) != OpusEncoder.FRAME_SIZE:
            # Close the file as soon as the playback is finished
            self._file.close()
            # Reopen the file, so it can be reused
            self._file = open(self.rpf.audio_filename, "rb")
            return b""
        return data

    def delete(self):
        """Permanently delete the downloaded file."""
        self._file.close()
        self.rpf.delete_audio_file()

    def __repr__(self):
        return f"<RoyalPCMAudio {self.rpf.audio_filename}>"



import os
import asyncio
import logging
from royalnet.bots import DiscordBot, DiscordConfig, TelegramBot, TelegramConfig
from royalnet.commands import *
from royalnet.commands.debug_create import DebugCreateCommand
from royalnet.commands.error_handler import ErrorHandlerCommand
from royalnet.network import RoyalnetServer, RoyalnetConfig
from royalnet.database import DatabaseConfig
from royalnet.database.tables import Royal, Telegram, Discord

loop = asyncio.get_event_loop()

log = logging.root
stream_handler = logging.StreamHandler()
stream_handler.formatter = logging.Formatter("{asctime}\t{name}\t{levelname}\t{message}", style="{")
log.addHandler(stream_handler)
log.setLevel(logging.WARNING)

commands = [PingCommand, ShipCommand, SmecdsCommand, ColorCommand, CiaoruoziCommand, DebugCreateCommand, SyncCommand,
            AuthorCommand, DiarioCommand, RageCommand, DateparserCommand, ReminderCommand, KvactiveCommand, KvCommand,
            KvrollCommand, VideoinfoCommand, SummonCommand, PlayCommand, SkipCommand, PlaymodeCommand,
            VideochannelCommand, CvCommand, PauseCommand, QueueCommand]

address, port = "127.0.0.1", 1234

print("Starting master...")
master = RoyalnetServer(address, port, os.environ["MASTER_KEY"])
loop.run_until_complete(master.start())

print("Starting bots...")
ds_bot = DiscordBot(discord_config=DiscordConfig(os.environ["DS_AK"]),
                    royalnet_config=RoyalnetConfig(f"ws://{address}:{port}", os.environ["MASTER_KEY"]),
                    database_config=DatabaseConfig(os.environ["DB_PATH"], Royal, Discord, "discord_id"),
                    commands=commands,
                    error_command=ErrorHandlerCommand,
                    missing_command=MissingCommand)
tg_bot = TelegramBot(telegram_config=TelegramConfig(os.environ["TG_AK"]),
                     royalnet_config=RoyalnetConfig(f"ws://{address}:{port}", os.environ["MASTER_KEY"]),
                     database_config=DatabaseConfig(os.environ["DB_PATH"], Royal, Telegram, "tg_id"),
                     commands=commands,
                     error_command=ErrorHandlerCommand,
                     missing_command=MissingCommand)
loop.create_task(tg_bot.run())
loop.create_task(ds_bot.run())

print("Running loop...")
loop.run_forever()



import asyncio
from ..utils import Command, Call
from ..error import InvalidInputError


class PingCommand(Command):

    command_name = "ping"
    command_description = "Ping pong dopo un po' di tempo!"
    command_syntax = "[time_to_wait]"

    @classmethod
    async def common(cls, call: Call):
        try:
            time = int(call.args[0])
        except InvalidInputError:
            time = 0
        except ValueError:
            raise InvalidInputError("time_to_wait is not a number")
        await asyncio.sleep(time)
        await call.reply("🏓 Pong!")



from sqlalchemy import Column, \
                       Integer, \
                       String, \
                       LargeBinary
from sqlalchemy.ext.declarative import declared_attr


class Royal:
    __tablename__ = "royals"

    @declared_attr
    def uid(self):
        return Column(Integer, unique=True, primary_key=True)

    @declared_attr
    def username(self):
        return Column(String, unique=True, nullable=False)

    @declared_attr
    def password(self):
        return Column(LargeBinary)

    @declared_attr
    def role(self):
        return Column(String, nullable=False)

    @declared_attr
    def avatar(self):
        return Column(LargeBinary)

    def __repr__(self):
        return f"<Royal {self.username}>"

    def __str__(self):
        return f"[c]royalnet:{self.username}[/c]"



import re
import datetime
import telegram
import typing
import os
import aiohttp
from ..utils import Command, Call
from ..error import InvalidInputError, InvalidConfigError, ExternalError
from ..database.tables import Royal, Diario, Alias
from ..utils import asyncify


# NOTE: Requires imgur api key for image upload, get one at https://apidocs.imgur.com
class DiarioCommand(Command):

    command_name = "diario"
    command_description = "Aggiungi una citazione al Diario."
    command_syntax = "[!] \"(testo)\" --[autore], [contesto]"

    require_alchemy_tables = {Royal, Diario, Alias}

    @classmethod
    async def _telegram_to_imgur(cls, photosizes: typing.List[telegram.PhotoSize], caption="") -> str:
        # Select the largest photo
        largest_photo = sorted(photosizes, key=lambda p: p.width * p.height)[-1]
        # Get the photo url
        photo_file: telegram.File = await asyncify(largest_photo.get_file)
        # Forward the url to imgur, as an upload
        try:
            imgur_api_key = os.environ["IMGUR_CLIENT_ID"]
        except KeyError:
            raise InvalidConfigError("Missing IMGUR_CLIENT_ID envvar, can't upload images to imgur.")
        async with aiohttp.request("post", "https://api.imgur.com/3/upload", data={
            "image": photo_file.file_path,
            "type": "URL",
            "title": "Diario image",
            "description": caption
        }, headers={
            "Authorization": f"Client-ID {imgur_api_key}"
        }) as request:
            response = await request.json()
            if not response["success"]:
                raise ExternalError("imgur returned an error in the image upload.")
            return response["data"]["link"]

    @classmethod
    async def common(cls, call: Call):
        # Find the creator of the quotes
        creator = await call.get_author(error_if_none=True)
        # Recreate the full sentence
        raw_text = " ".join(call.args)
        # Pass the sentence through the diario regex
        match = re.match(r'(!)? *["«‘“‛‟❛❝〝＂`]([^"]+)["»’”❜❞〞＂`] *(?:(?:-{1,2}|—) *([\w ]+))?(?:, *([^ ].*))?', raw_text)
        # Find the corresponding matches
        if match is not None:
            spoiler = bool(match.group(1))
            text = match.group(2)
            quoted = match.group(3)
            context = match.group(4)
        # Otherwise, consider everything part of the text
        else:
            spoiler = False
            text = raw_text
            quoted = None
            context = None
        timestamp = datetime.datetime.now()
        # Ensure there is some text
        if not text:
            raise InvalidInputError("Missing text.")
        # Or a quoted
        if not quoted:
            quoted = None
        if not context:
            context = None
        # Find if there's a Royalnet account associated with the quoted name
        if quoted is not None:
            quoted_alias = await asyncify(call.session.query(call.alchemy.Alias).filter_by(alias=quoted.lower()).one_or_none)
        else:
            quoted_alias = None
        quoted_account = quoted_alias.royal if quoted_alias is not None else None
        if quoted_alias is not None and quoted_account is None:
            await call.reply("⚠️ Il nome dell'autore è ambiguo, quindi la riga non è stata aggiunta.\nPer piacere, ripeti il comando con un nome più specifico!")
            return
        # Create the diario quote
        diario = call.alchemy.Diario(creator=creator,
                                     quoted_account=quoted_account,
                                     quoted=quoted,
                                     text=text,
                                     context=context,
                                     timestamp=timestamp,
                                     media_url=None,
                                     spoiler=spoiler)
        call.session.add(diario)
        await asyncify(call.session.commit)
        await call.reply(f"✅ {str(diario)}")

    @classmethod
    async def telegram(cls, call: Call):
        update: telegram.Update = call.kwargs["update"]
        message: telegram.Message = update.message
        reply: telegram.Message = message.reply_to_message
        creator = await call.get_author()
        # noinspection PyUnusedLocal
        quoted_account: typing.Optional[call.alchemy.Telegram]
        # noinspection PyUnusedLocal
        quoted: typing.Optional[str]
        # noinspection PyUnusedLocal
        text: typing.Optional[str]
        # noinspection PyUnusedLocal
        context: typing.Optional[str]
        # noinspection PyUnusedLocal
        timestamp: datetime.datetime
        # noinspection PyUnusedLocal
        media_url: typing.Optional[str]
        # noinspection PyUnusedLocal
        spoiler: bool
        if creator is None:
            await call.reply("⚠️ Devi essere registrato a Royalnet per usare questo comando!")
            return
        if reply is not None:
            # Get the message text
            text = reply.text
            # Check if there's an image associated with the reply
            photosizes: typing.Optional[typing.List[telegram.PhotoSize]] = reply.photo
            if photosizes:
                # Text is a caption
                text = reply.caption
                media_url = await cls._telegram_to_imgur(photosizes, text if text is not None else "")
            else:
                media_url = None
            # Ensure there is a text or an image
            if not (text or media_url):
                raise InvalidInputError("Missing text.")
            # Find the Royalnet account associated with the sender
            quoted_tg = await asyncify(call.session.query(call.alchemy.Telegram).filter_by(tg_id=reply.from_user.id).one_or_none)
            quoted_account = quoted_tg.royal if quoted_tg is not None else None
            # Find the quoted name to assign
            quoted_user: telegram.User = reply.from_user
            quoted = quoted_user.full_name
            # Get the timestamp
            timestamp = reply.date
            # Set the other properties
            spoiler = False
            context = None
        else:
            # Get the current timestamp
            timestamp = datetime.datetime.now()
            # Get the message text
            raw_text = " ".join(call.args)
            # Check if there's an image associated with the reply
            photosizes: typing.Optional[typing.List[telegram.PhotoSize]] = message.photo
            if photosizes:
                media_url = await cls._telegram_to_imgur(photosizes, raw_text if raw_text is not None else "")
            else:
                media_url = None
            # Parse the text, if it exists
            if raw_text:
                # Pass the sentence through the diario regex
                match = re.match(r'(!)? *["«‘“‛‟❛❝〝＂`]([^"]+)["»’”❜❞〞＂`] *(?:(?:-{1,2}|—) *([\w ]+))?(?:, *([^ ].*))?',
                                 raw_text)
                # Find the corresponding matches
                if match is not None:
                    spoiler = bool(match.group(1))
                    text = match.group(2)
                    quoted = match.group(3)
                    context = match.group(4)
                # Otherwise, consider everything part of the text
                else:
                    spoiler = False
                    text = raw_text
                    quoted = None
                    context = None
                # Ensure there's a quoted
                if not quoted:
                    quoted = None
                if not context:
                    context = None
                # Find if there's a Royalnet account associated with the quoted name
                if quoted is not None:
                    quoted_alias = await asyncify(
                        call.session.query(call.alchemy.Alias).filter_by(alias=quoted.lower()).one_or_none)
                else:
                    quoted_alias = None
                quoted_account = quoted_alias.royal if quoted_alias is not None else None
            else:
                text = None
                quoted = None
                quoted_account = None
                spoiler = False
                context = None
            # Ensure there is a text or an image
            if not (text or media_url):
                raise InvalidInputError("Missing text.")
        # Create the diario quote
        diario = call.alchemy.Diario(creator=creator,
                                     quoted_account=quoted_account,
                                     quoted=quoted,
                                     text=text,
                                     context=context,
                                     timestamp=timestamp,
                                     media_url=media_url,
                                     spoiler=spoiler)
        call.session.add(diario)
        await asyncify(call.session.commit)
        await call.reply(f"✅ {str(diario)}")



import asyncio
import functools
import typing


async def asyncify(function: typing.Callable, *args, **kwargs):
    """Convert a function into a coroutine.

    Warning:
        The coroutine cannot be cancelled, and any attempts to do so will result in unexpected outputs."""
    loop = asyncio.get_event_loop()
    return await loop.run_in_executor(None, functools.partial(function, *args, **kwargs))



def discord_escape(string: str) -> str:
    """Escape a string to be sent through Discord, and format it using RoyalCode.
    
    Warning:
        Currently escapes everything, even items in code blocks."""
    return string.replace("*", "\\*") \
                 .replace("_", "\\_") \
                 .replace("`", "\\`") \
                 .replace("[b]", "**") \
                 .replace("[/b]", "**") \
                 .replace("[i]", "_") \
                 .replace("[/i]", "_") \
                 .replace("[u]", "__") \
                 .replace("[/u]", "__") \
                 .replace("[c]", "`") \
                 .replace("[/c]", "`") \
                 .replace("[p]", "```") \
                 .replace("[/p]", "```")


def telegram_escape(string: str) -> str:
    """Escape a string to be sent through Telegram, and format it using RoyalCode.
    
    Warning:
        Currently escapes everything, even items in code blocks."""
    return string.replace("<", "&lt;") \
                 .replace(">", "&gt;") \
                 .replace("[b]", "<b>") \
                 .replace("[/b]", "</b>") \
                 .replace("[i]", "<i>") \
                 .replace("[/i]", "</i>") \
                 .replace("[u]", "<b>") \
                 .replace("[/u]", "</b>") \
                 .replace("[c]", "<code>") \
                 .replace("[/c]", "</code>") \
                 .replace("[p]", "<pre>") \
                 .replace("[/p]", "</pre>")



import telegram
import telegram.utils.request
import asyncio
import typing
import logging as _logging
from .generic import GenericBot
from ..commands import NullCommand
from ..utils import asyncify, Call, Command, telegram_escape
from ..error import UnregisteredError, InvalidConfigError, RoyalnetResponseError
from ..network import RoyalnetConfig, Request, ResponseSuccess, ResponseError
from ..database import DatabaseConfig

loop = asyncio.get_event_loop()
log = _logging.getLogger(__name__)


class TelegramConfig:
    """The specific configuration to be used for :py:class:`royalnet.database.TelegramBot`."""
    def __init__(self, token: str):
        self.token: str = token


class TelegramBot(GenericBot):
    """A bot that connects to `Telegram <https://telegram.org/>`_."""
    interface_name = "telegram"

    def _init_client(self):
        """Create the :py:class:`telegram.Bot`, and set the starting offset."""
        # https://github.com/python-telegram-bot/python-telegram-bot/issues/341
        request = telegram.utils.request.Request(5)
        self.client = telegram.Bot(self._telegram_config.token, request=request)
        self._offset: int = -100

    def _call_factory(self) -> typing.Type[Call]:
        # noinspection PyMethodParameters
        class TelegramCall(Call):
            interface_name = self.interface_name
            interface_obj = self
            interface_prefix = "/"

            alchemy = self.alchemy

            async def reply(call, text: str):
                await asyncify(call.channel.send_message, telegram_escape(text), parse_mode="HTML")

            async def net_request(call, request: Request, destination: str) -> dict:
                if self.network is None:
                    raise InvalidConfigError("Royalnet is not enabled on this bot")
                response_dict: dict = await self.network.request(request.to_dict(), destination)
                if "type" not in response_dict:
                    raise RoyalnetResponseError("Response is missing a type")
                elif response_dict["type"] == "ResponseSuccess":
                    response: typing.Union[ResponseSuccess, ResponseError] = ResponseSuccess.from_dict(response_dict)
                elif response_dict["type"] == "ResponseError":
                    response = ResponseError.from_dict(response_dict)
                else:
                    raise RoyalnetResponseError("Response type is unknown")
                response.raise_on_error()
                return response.data

            async def get_author(call, error_if_none=False):
                update: telegram.Update = call.kwargs["update"]
                user: telegram.User = update.effective_user
                if user is None:
                    if error_if_none:
                        raise UnregisteredError("No author for this message")
                    return None
                query = call.session.query(self.master_table)
                for link in self.identity_chain:
                    query = query.join(link.mapper.class_)
                query = query.filter(self.identity_column == user.id)
                result = await asyncify(query.one_or_none)
                if result is None and error_if_none:
                    raise UnregisteredError("Author is not registered")
                return result

        return TelegramCall

    def __init__(self, *,
                 telegram_config: TelegramConfig,
                 royalnet_config: typing.Optional[RoyalnetConfig] = None,
                 database_config: typing.Optional[DatabaseConfig] = None,
                 command_prefix: str = "/",
                 commands: typing.List[typing.Type[Command]] = None,
                 missing_command: typing.Type[Command] = NullCommand,
                 error_command: typing.Type[Command] = NullCommand):
        super().__init__(royalnet_config=royalnet_config,
                         database_config=database_config,
                         command_prefix=command_prefix,
                         commands=commands,
                         missing_command=missing_command,
                         error_command=error_command)
        self._telegram_config = telegram_config
        self._init_client()

    async def _handle_update(self, update: telegram.Update):
        # Skip non-message updates
        if update.message is None:
            return
        message: telegram.Message = update.message
        text: str = message.text
        # Try getting the caption instead
        if text is None:
            text: str = message.caption
        # No text or caption, ignore the message
        if text is None:
            return
        # Find and clean parameters
        command_text, *parameters = text.split(" ")
        command_name = command_text.replace(f"@{self.client.username}", "").lower()
        # Call the command
        await self.call(command_name, update.message.chat, parameters, update=update)

    async def run(self):
        while True:
            # Get the latest 100 updates
            try:
                last_updates: typing.List[telegram.Update] = await asyncify(self.client.get_updates, offset=self._offset, timeout=60)
            except telegram.error.TimedOut:
                continue
            # Handle updates
            for update in last_updates:
                # noinspection PyAsyncCall
                loop.create_task(self._handle_update(update))
            # Recalculate offset
            try:
                self._offset = last_updates[-1].update_id + 1
            except IndexError:
                pass

    @property
    def botfather_command_string(self) -> str:
        """Generate a string to be pasted in the "Edit Commands" BotFather prompt."""
        string = ""
        for command_key in self.commands:
            command = self.commands[command_key]
            string += f"{command.command_name} - {command.command_description}\n"
        return string



import flask as f
import os
from ... import Royalprint


tmpl_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'templates')
bp = Royalprint("home", __name__, template_folder="templates")


@bp.route("/")
def home_index():
    return f.render_template("home.html")



from sqlalchemy import Column, \
                       String, \
                       ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declared_attr
# noinspection PyUnresolvedReferences
from .keygroups import Keygroup


class Keyvalue:
    __tablename__ = "keyvalues"

    @declared_attr
    def group_name(self):
        return Column(String, ForeignKey("keygroups.group_name"), primary_key=True)

    @declared_attr
    def key(self):
        return Column(String, primary_key=True)

    @declared_attr
    def value(self):
        return Column(String, nullable=False)

    @declared_attr
    def group(self):
        return relationship("Keygroup")

    def __repr__(self):
        return f"<Keyvalue group={self.group_name} key={self.key} value={self.value}>"

    def __str__(self):
        return f"{self.key}: [b]{self.value}[/b]"



"""Video and audio downloading related classes, mainly used for Discord voice bots."""

from .playmodes import PlayMode, Playlist, Pool
from .youtubedl import YtdlFile, YtdlInfo
from .royalpcmfile import RoyalPCMFile
from .royalpcmaudio import RoyalPCMAudio

__all__ = ["PlayMode", "Playlist", "Pool", "YtdlFile", "YtdlInfo", "RoyalPCMFile", "RoyalPCMAudio"]



import os
import re
import datetime
import difflib
import uuid
from royalnet.database import Alchemy
from royalnet.database.tables import Royal, WikiPage, WikiRevision


if __name__ == "__main__":
    alchemy = Alchemy(os.environ["DB_PATH"], {Royal, WikiPage, WikiRevision})
    with open(r"data.txt") as file, alchemy.session_cm() as session:
        for line in file.readlines():
            match = re.match("^([^\t]+)\t([^\t]+)\t([tf])$", line)
            if match is None:
                continue
            title = match.group(1)
            content = match.group(2).replace(r"\r\n", "\n").replace(r"\t", "\t")
            page = alchemy.WikiPage(page_id=uuid.uuid4(),
                                    title=title,
                                    content=content)
            session.flush()
            revision = alchemy.WikiRevision(revision_id=uuid.uuid4(),
                                            page=page,
                                            author_id=31,  # Royalbot
                                            timestamp=datetime.datetime.now(),
                                            reason="Imported from 'four' database",
                                            diff="\n".join(difflib.unified_diff([], content.split("\n"))))
            session.add(page)
            session.add(revision)
            print(f"{title} done.")
        session.commit()



from ..utils import Command, Call
from ..database.tables import Royal, Telegram


class AuthorCommand(Command):

    command_name = "author"
    command_description = "Ottieni informazioni sull'autore di questa chiamata."
    command_syntax = ""

    require_alchemy_tables = {Royal, Telegram}

    @classmethod
    async def common(cls, call: Call):
        author = await call.get_author()
        if author is None:
            await call.reply(f"☁️ L'autore di questa chiamata è sconosciuto.")
        await call.reply(f"🌞 {str(author)} è l'autore di questa chiamata.")



"""Miscellaneous useful functions and classes."""

from .asyncify import asyncify
from .escaping import telegram_escape, discord_escape
from .call import Call
from .command import Command
from .commandargs import CommandArgs
from .safeformat import safeformat
from .classdictjanitor import cdj
from .sleepuntil import sleep_until
from .networkhandler import NetworkHandler
from .formatters import andformat, plusformat, fileformat, ytdldateformat, numberemojiformat

__all__ = ["asyncify", "Call", "Command", "safeformat", "cdj", "sleep_until", "plusformat", "CommandArgs",
           "NetworkHandler", "andformat", "plusformat", "fileformat", "ytdldateformat", "numberemojiformat"]



import typing
if typing.TYPE_CHECKING:
    from .network import ResponseError


class NoneFoundError(Exception):
    """The element that was being looked for was not found."""


class TooManyFoundError(Exception):
    """Multiple elements matching the request were found, and only one was expected."""


class UnregisteredError(Exception):
    """The command required a registered user, and the user was not registered."""


class UnsupportedError(Exception):
    """The command is not supported for the specified interface."""


class InvalidInputError(Exception):
    """The command has received invalid input and cannot complete."""


class InvalidConfigError(Exception):
    """The bot has not been configured correctly, therefore the command can not function."""


class RoyalnetRequestError(Exception):
    """An error was raised while handling the Royalnet request.

    This exception contains the :py:class:`royalnet.network.ResponseError` that was returned by the other Link."""
    def __init__(self, error: "ResponseError"):
        self.error: "ResponseError" = error


class RoyalnetResponseError(Exception):
    """The :py:class:`royalnet.network.Response` that was received is invalid."""


class ExternalError(Exception):
    """Something went wrong in a non-Royalnet component and the command execution cannot be completed."""


class FileTooBigError(Exception):
    """The file to be downloaded would be too big to store; therefore, it has been skipped."""



import typing
import flask as f


class Royalprint(f.Blueprint):
    def __init__(self, name, import_name, static_folder=None,
                 static_url_path=None, template_folder=None,
                 url_prefix=None, subdomain=None, url_defaults=None,
                 root_path=None, required_tables: typing.Optional[set] = None):
        super().__init__(name, import_name, static_folder=static_folder,
                         static_url_path=static_url_path, template_folder=template_folder,
                         url_prefix=url_prefix, subdomain=subdomain, url_defaults=url_defaults,
                         root_path=root_path)
        self.required_tables = required_tables
        if self.required_tables is None:
            self.required_tables = set()



from ..database.tables import ActiveKvGroup, Royal, Keyvalue, Keygroup
from ..utils import Command, Call, asyncify


class KvCommand(Command):

    command_name = "kv"
    command_description = "Visualizza o modifica un valore kv."
    command_syntax = "(chiave) [valore]"

    require_alchemy_tables = {ActiveKvGroup, Royal, Keyvalue, Keygroup}

    @classmethod
    async def common(cls, call: Call):
        key = call.args[0].lower()
        value = call.args.optional(1)
        author = await call.get_author(error_if_none=True)
        active = await asyncify(call.session.query(call.alchemy.ActiveKvGroup).filter_by(royal=author).one_or_none)
        if active is None:
            await call.reply("⚠️ Devi prima attivare un gruppo con il comando [c]kvactive[/c]!")
            return
        keyvalue = await asyncify(call.session.query(call.alchemy.Keyvalue).filter_by(group=active.group, key=key).one_or_none)
        if value is None:
            # Get
            if keyvalue is None:
                await call.reply("⚠️ La chiave specificata non esiste.")
                return
            await call.reply(f"ℹ️ Valore della chiave:\n{keyvalue}")
        else:
            # Set/kv asdf 1000
            if keyvalue is None:
                keyvalue = call.alchemy.Keyvalue(group=active.group, key=key, value=value)
                call.session.add(keyvalue)
            else:
                keyvalue.value = value
            await asyncify(call.session.commit)
            await call.reply(f"✅ Chiave aggiornata:\n{keyvalue}")



class Request:
    """A request sent from a :py:class:`royalnet.network.RoyalnetLink` to another.

     It contains the name of the requested handler, in addition to the data."""

    def __init__(self, handler: str, data: dict):
        super().__init__()
        self.handler: str = handler
        self.data: dict = data

    def to_dict(self):
        return self.__dict__

    @staticmethod
    def from_dict(d: dict):
        return Request(**d)

    def __eq__(self, other):
        if isinstance(other, Request):
            return self.handler == other.handler and self.data == other.data
        return False

    def __repr__(self):
        return f"royalnet.network.Request(handler={self.handler}, data={self.data})"



"""Royalnet (websocket) related classes."""
from .request import Request
from .response import Response, ResponseSuccess, ResponseError
from .package import Package
from .royalnetlink import RoyalnetLink, NetworkError, NotConnectedError, NotIdentifiedError, ConnectionClosedError
from .royalnetserver import RoyalnetServer
from .royalnetconfig import RoyalnetConfig

__all__ = ["RoyalnetLink",
           "NetworkError",
           "NotConnectedError",
           "NotIdentifiedError",
           "Package",
           "RoyalnetServer",
           "RoyalnetConfig",
           "ConnectionClosedError",
           "Request",
           "Response",
           "ResponseSuccess",
           "ResponseError"]
